@@ -3,7 +3,7 @@ filesystem key/value storage with support for storing and retrieving
 basic python types based on parametrizable keys.  Multiple
 read Transactions can execute concurrently while at most one
 write Transaction is ongoing.  Each Transaction will see a consistent
-view of key/values refering to the point in time it was started,
+view of key/values referring to the point in time it was started,
 independent from any future changes.
 """
 from __future__ import unicode_literals
@@ -11,16 +11,19 @@ import re
 import contextlib
 import py
 from . import mythread
+from .fileutil import loads
 from .log import threadlog, thread_push_log, thread_pop_log
 from .readonly import get_mutable_deepcopy, ensure_deeply_readonly, \
                       is_deeply_readonly
 from .fileutil import read_int_from_file, write_int_to_file
-
+import attr
 import time
 
 from devpi_common.types import cached_property
 
 
+notset = object()
+
 
 class TxNotificationThread:
     def __init__(self, keyfs):
@@ -43,7 +46,6 @@ class TxNotificationThread:
                 while serial > self.read_event_serial():
                     self.cv_new_event_serial.wait()
 
-
     def read_event_serial(self):
         # the disk serial is kept one higher because pre-2.1.2
         # "event_serial" pointed to the "next event serial to be
@@ -131,9 +133,12 @@ class KeyFS(object):
             notify_on_commit=self._notify_on_commit,
             cache_size=cache_size)
         self._readonly = readonly
+
+    def finalize_init(self):
         self._storage.perform_crash_recovery()
 
     def import_changes(self, serial, changes):
+        subscriber_task_infos = []
         with self._storage.get_connection(write=True) as conn:
             with conn.write_transaction() as fswriter:
                 next_serial = conn.last_changelog_serial + 1
@@ -141,11 +146,25 @@ class KeyFS(object):
                 for relpath, tup in changes.items():
                     keyname, back_serial, val = tup
                     typedkey = self.get_key_instance(keyname, relpath)
-                    fswriter.record_set(typedkey, get_mutable_deepcopy(val))
+                    try:
+                        fswriter.record_set(
+                            typedkey, get_mutable_deepcopy(val),
+                            back_serial=back_serial)
+                    except TypeError as e:
+                        # for backward compatibility with storage backends
+                        # which didn't support the back_serial keyword
+                        if not len(e.args) or 'back_serial' not in e.args[0]:
+                            raise
+                        fswriter.record_set(
+                            typedkey, get_mutable_deepcopy(val))
                     meth = self._import_subscriber.get(keyname)
                     if meth is not None:
-                        threadlog.debug("calling import subscriber %r", meth)
-                        with self.transaction(write=False, at_serial=serial):
+                        subscriber_task_infos.append(
+                            (meth, typedkey, val, back_serial))
+                if subscriber_task_infos:
+                    with self.transaction(write=False, at_serial=serial):
+                        for meth, typedkey, val, back_serial in subscriber_task_infos:
+                            threadlog.debug("calling import subscriber %r", meth)
                             meth(fswriter, typedkey, val, back_serial)
 
     def subscribe_on_import(self, key, subscriber):
@@ -206,8 +225,12 @@ class KeyFS(object):
             key = PTypedKey(self, path, type, name)
         else:
             key = TypedKey(self, path, type, name)
+        if name in self._keys:
+            raise ValueError("Duplicate registration for key named '%s'" % name)
         self._keys[name] = key
         setattr(self, name, key)
+        if hasattr(self._storage, 'add_key'):
+            self._storage.add_key(key)
         return key
 
     def get_key(self, name):
@@ -272,11 +295,13 @@ class KeyFS(object):
 
 class PTypedKey:
     rex_braces = re.compile(r'\{(.+?)\}')
+
     def __init__(self, keyfs, key, type, name):
         self.keyfs = keyfs
         self.pattern = py.builtin._totext(key)
         self.type = type
         self.name = name
+
         def repl(match):
             name = match.group(1)
             return r'(?P<%s>[^\/]+)' % name
@@ -369,6 +394,15 @@ class TypedKey:
         return self.keyfs.tx.delete(self)
 
 
+@attr.s(slots=True)
+class RelpathInfo(object):
+    relpath = attr.ib(type=str)
+    keyname = attr.ib(type=str)
+    serial = attr.ib(type=int)
+    back_serial = attr.ib(type=int)
+    value = attr.ib()
+
+
 class Transaction(object):
     def __init__(self, keyfs, at_serial=None, write=False):
         self.keyfs = keyfs
@@ -378,35 +412,74 @@ class Transaction(object):
         if at_serial is None:
             at_serial = self.conn.last_changelog_serial
         self.at_serial = at_serial
+        self._original = {}
         self.cache = {}
         self.dirty = set()
         self.closed = False
         self.doomed = False
 
-    def _get_value_at(self, typedkey, at_serial):
-        relpath = typedkey.relpath
-        keyname, last_serial = self.conn.db_read_typedkey(relpath)
+    def iter_relpaths_at(self, typedkeys, at_serial):
+        keynames = frozenset(k.name for k in typedkeys)
+        seen = set()
+        for serial in range(at_serial, -1, -1):
+            raw_entry = self.conn.get_raw_changelog_entry(serial)
+            changes = loads(raw_entry)[0]
+            for relpath, (keyname, back_serial, val) in changes.items():
+                if keyname not in keynames:
+                    continue
+                if relpath not in seen:
+                    seen.add(relpath)
+                    yield RelpathInfo(
+                        relpath=relpath, keyname=keyname,
+                        serial=serial, back_serial=back_serial,
+                        value=val)
+
+    def iter_serial_and_value_backwards(self, relpath, last_serial):
         while last_serial >= 0:
             tup = self.conn.get_changes(last_serial).get(relpath)
-            assert tup, "no transaction entry at %s" %(last_serial)
+            if tup is None:
+                raise RuntimeError("no transaction entry at %s" % (last_serial))
             keyname, back_serial, val = tup
-            if last_serial > at_serial:
-                last_serial = back_serial
-                continue
-            if val is not None:
-                return (last_serial, val)
-            raise KeyError(relpath)  # was deleted
+            yield (last_serial, val)
+            last_serial = back_serial
 
         # we could not find any change below at_serial which means
         # the key didn't exist at that point in time
-        raise KeyError(relpath)
+        return
+
+    def get_last_serial_and_value_at(self, typedkey, at_serial, raise_on_error=True):
+        relpath = typedkey.relpath
+        try:
+            (keyname, last_serial) = self.conn.db_read_typedkey(relpath)
+        except KeyError:
+            if raise_on_error:
+                raise
+            return None
+        serials_and_values = self.iter_serial_and_value_backwards(
+            relpath, last_serial)
+        try:
+            (last_serial, val) = next(serials_and_values)
+            while last_serial >= 0:
+                if last_serial > at_serial:
+                    (last_serial, val) = next(serials_and_values)
+                    continue
+                if val is not None or not raise_on_error:
+                    return (last_serial, val)
+                raise KeyError(relpath)  # was deleted
+        except StopIteration:
+            pass
+
+        if raise_on_error:
+            # we could not find any change below at_serial which means
+            # the key didn't exist at that point in time
+            raise KeyError(relpath)
 
     def get_value_at(self, typedkey, at_serial):
-        (last_serial, val) = self._get_value_at(typedkey, at_serial)
+        (last_serial, val) = self.get_last_serial_and_value_at(typedkey, at_serial)
         return val
 
     def last_serial(self, typedkey):
-        (last_serial, val) = self._get_value_at(typedkey, self.at_serial)
+        (last_serial, val) = self.get_last_serial_and_value_at(typedkey, self.at_serial)
         return last_serial
 
     def derive_key(self, relpath):
@@ -429,16 +502,28 @@ class Transaction(object):
     def is_dirty(self, typedkey):
         return typedkey in self.dirty
 
+    def get_original(self, typedkey):
+        """ Return original value from start of transaction,
+            without changes from current transaction."""
+        try:
+            return self._original[typedkey]
+        except KeyError:
+            # will raise KeyError if it doesn't exist
+            val = self.get_value_at(typedkey, self.at_serial)
+            assert is_deeply_readonly(val)
+            self._original[typedkey] = val
+        return val
+
     def get(self, typedkey, readonly=True):
-        """ Return value referenced by typedkey, either as a readonly-view
-        or as a mutable deep copy. """
+        """ Return current value referenced by typedkey,
+            either as a readonly-view or as a mutable deep copy. """
         try:
             val = self.cache[typedkey]
         except KeyError:
             absent = typedkey in self.dirty
             if not absent:
                 try:
-                    val = self.get_value_at(typedkey, self.at_serial)
+                    val = self.get_original(typedkey)
                 except KeyError:
                     absent = True
             if absent:
@@ -459,10 +544,12 @@ class Transaction(object):
         if typedkey in self.dirty:
             return False
         try:
-            self.get_value_at(typedkey, self.at_serial)
+            val = self.get_value_at(typedkey, self.at_serial)
         except KeyError:
             return False
         else:
+            assert is_deeply_readonly(val)
+            self.cache[typedkey] = val
             return True
 
     def delete(self, typedkey):
@@ -478,8 +565,15 @@ class Transaction(object):
         # keys, not bytes
         if typedkey.type == dict:
             check_unicode_keys(val)
+        try:
+            old_val = self.get_original(typedkey)
+        except KeyError:
+            old_val = notset
         self.cache[typedkey] = val
-        self.dirty.add(typedkey)
+        if val != old_val:
+            self.dirty.add(typedkey)
+        else:
+            self.dirty.discard(typedkey)
 
     def commit(self):
         if self.doomed:
@@ -517,6 +611,8 @@ class Transaction(object):
         return self.at_serial
 
     def rollback(self):
+        if hasattr(self.conn, 'rollback'):
+            self.conn.rollback()
         threadlog.debug("transaction rollback at %s" % (self.at_serial))
         return self._close()
 
