@@ -1,14 +1,19 @@
 from __future__ import unicode_literals
 
-import collections
 import os
 import py
 import re
+import traceback
 from time import time
+try:
+    from collections.abc import Iterator
+except ImportError:
+    from collections import Iterator
 from devpi_common.types import ensure_unicode
 from devpi_common.url import URL
 from devpi_common.metadata import get_pyversion_filetype
 import devpi_server
+from io import BytesIO
 from pluggy import HookimplMarker
 from pyramid.compat import escape
 from pyramid.compat import urlparse
@@ -18,6 +23,7 @@ from pyramid.httpexceptions import HTTPUnauthorized
 from pyramid.httpexceptions import exception_response
 from pyramid.response import Response
 from pyramid.security import forget
+from pyramid.view import exception_view_config
 from pyramid.view import view_config
 import itertools
 import json
@@ -26,8 +32,9 @@ from devpi_common.validation import normalize_name, is_valid_archive_name
 
 from .filestore import BadGateway
 from .model import InvalidIndex, InvalidIndexconfig, InvalidUser
+from .model import ReadonlyIndex
+from .model import RemoveValue
 from .model import UpstreamError
-from .model import get_ixconfigattrs
 from .readonly import get_mutable_deepcopy
 from .log import thread_push_log, thread_pop_log, threadlog
 
@@ -92,13 +99,14 @@ class HTTPResponse(HTTPSuccessful):
     body_template = None
     comment = None
     detail = None
+
     def __init__(self, **kw):
         Response.__init__(self, **kw)
         Exception.__init__(self)
 
 
 def apireturn(code, message=None, result=None, type=None):
-    d = dict() # status=code)
+    d = dict()
     if result is not None:
         assert type is not None
         d["result"] = result
@@ -199,8 +207,9 @@ def make_uuid_headers(nodeinfo):
 def tween_keyfs_transaction(handler, registry):
     keyfs = registry["xom"].keyfs
     is_replica = registry["xom"].is_replica()
+
     def request_tx_handler(request):
-        write  = is_mutating_http_method(request.method) and not is_replica
+        write = is_mutating_http_method(request.method) and not is_replica
         with keyfs.transaction(write=write) as tx:
             threadlog.debug("in-transaction %s", tx.at_serial)
             response = handler(request)
@@ -210,7 +219,7 @@ def tween_keyfs_transaction(handler, registry):
 
 
 def set_header_devpi_serial(response, tx):
-    if isinstance(response._app_iter, collections.Iterator):
+    if isinstance(response._app_iter, Iterator):
         return
     if tx.commit_serial is not None:
         serial = tx.commit_serial
@@ -231,7 +240,10 @@ def get_actions(json):
             op = 'add'
             key = key[:-1]
         elif key.endswith('-'):
-            op = 'del'
+            if value == '':
+                op = 'drop'
+            else:
+                op = 'del'
             key = key[:-1]
         else:
             op = 'set'
@@ -239,6 +251,13 @@ def get_actions(json):
     return result
 
 
+@exception_view_config(ReadonlyIndex)
+def readonly_index_view(exc, request):
+    response = Response("%s" % exc)
+    response.status = "405 %s" % exc
+    return response
+
+
 class StatusView:
     def __init__(self, request):
         self.request = request
@@ -251,7 +270,7 @@ class StatusView:
             "serverdir": str(config.serverdir),
             "uuid": self.xom.config.nodeinfo["uuid"],
             "versioninfo":
-                    dict(self.request.registry["devpi_version_info"]),
+                dict(self.request.registry["devpi_version_info"]),
             "server-code": os.path.dirname(devpi_server.__file__),
             "host": config.args.host,
             "port": config.args.port,
@@ -263,13 +282,23 @@ class StatusView:
                 self.xom.keyfs.notifier.get_event_serial_timestamp(),
             "event-serial-in-sync-at":
                 self.xom.keyfs.notifier.event_serial_in_sync_at,
+            "metrics": [],
         }
-        master_url = config.args.master_url
-        if master_url:
+        seen = set()
+        for results in config.hook.devpiserver_metrics(request=self.request):
+            for metric in results:
+                name = metric[0]
+                if name in seen:
+                    raise RuntimeError(
+                        "A plugin returned a metric with the name '%s' which "
+                        "is already used." % name)
+                seen.add(name)
+                status["metrics"].append(metric)
+        if self.xom.is_replica():
             from .replica import ReplicationErrors
             status["role"] = "REPLICA"
-            status["master-url"] = master_url
-            status["master-uuid"] = config.nodeinfo.get("master-uuid")
+            status["master-url"] = config.master_url.url
+            status["master-uuid"] = config.get_master_uuid()
             status["master-serial"] = self.xom.replica_thread.get_master_serial()
             status["master-serial-timestamp"] = self.xom.replica_thread.get_master_serial_timestamp()
             status["replica-started-at"] = self.xom.replica_thread.started_at
@@ -388,7 +417,7 @@ class PyPIView:
                     "index": request.stage_url(stage),
                     "simpleindex": request.simpleindex_url(stage)
                 })
-                if stage.ixconfig["type"] == "stage":
+                if stage.ixconfig["type"] != "mirror":
                     api["pypisubmit"] = request.route_url(
                         "/{user}/{index}/", user=user, index=index)
         apireturn(200, type="apiconfig", result=api)
@@ -553,7 +582,7 @@ class PyPIView:
         self.log.info("starting +simple")
         stage = self.context.stage
         try:
-            stage_results = list(stage.op_sro("list_projects_perstage"))
+            stage_results = list(stage.list_projects())
         except stage.UpstreamError as e:
             threadlog.error(e.msg)
             abort(self.request, 502, e.msg)
@@ -564,8 +593,7 @@ class PyPIView:
     def _simple_list_all(self, stage, stage_results):
         response = self.request.response
         response.content_type = "text/html ; charset=utf-8"
-        title =  "%s: simple list (including inherited indices)" %(
-                 stage.name)
+        title = "%s: simple list (including inherited indices)" % (stage.name)
         yield ("<!DOCTYPE html><html><head><title>%s</title></head><body><h1>%s</h1>" %(
               title, title)).encode("utf-8")
         all_names = set()
@@ -620,14 +648,27 @@ class PyPIView:
         if not self.request.has_permission("index_create"):
             apireturn(403, "no permission to create index %s/%s" % (
                 self.context.username, self.context.index))
-        kvdict = getkvdict_index(self.xom.config.hook, getjson(self.request))
+        json = getjson(self.request)
         try:
-            stage = self.context.user.create_stage(self.context.index, **kvdict)
+            stage = self.context.user.create_stage(self.context.index, **json)
             ixconfig = stage.ixconfig
         except InvalidIndex as e:
             apireturn(400, "%s" % e)
         except InvalidIndexconfig as e:
             apireturn(400, message=", ".join(e.messages))
+        try:
+            stage.customizer.on_modified(self.request, {})
+        except InvalidIndexconfig as e:
+            self.request.apifatal(400, message=", ".join(e.messages))
+        except HTTPException:
+            if not self.request.registry["xom"].keyfs.tx.doomed:
+                self.request.apifatal(
+                    500,
+                    "An HTTPException was raised. In on_modified this is not "
+                    "allowed, as it breaks transaction handling. Use "
+                    "request.apifatal instead.")
+            else:
+                raise
         apireturn(200, type="indexconfig", result=ixconfig)
 
     @view_config(
@@ -655,12 +696,29 @@ class PyPIView:
                         ixconfig[key].append(value)
                 elif op == 'set':
                     ixconfig[key] = value
+                elif op == 'drop':
+                    ixconfig[key] = RemoveValue
+                else:
+                    raise ValueError("Unknown operator '%s'." % op)
             json = ixconfig
-        kvdict = getkvdict_index(self.xom.config.hook, json)
+        oldconfig = dict(stage.ixconfig)
         try:
-            ixconfig = stage.modify(**kvdict)
+            ixconfig = stage.modify(**json)
         except InvalidIndexconfig as e:
             apireturn(400, message=", ".join(e.messages))
+        try:
+            stage.customizer.on_modified(self.request, oldconfig)
+        except InvalidIndexconfig as e:
+            self.request.apifatal(400, message=", ".join(e.messages))
+        except HTTPException:
+            if not self.request.registry["xom"].keyfs.tx.doomed:
+                self.request.apifatal(
+                    500,
+                    "An HTTPException was raised. In on_modified this is not "
+                    "allowed, as it breaks transaction handling. Use "
+                    "request.apifatal instead.")
+            else:
+                raise
         apireturn(200, type="indexconfig", result=ixconfig)
 
     @view_config(
@@ -715,7 +773,7 @@ class PyPIView:
             self.log.debug("targetindex %r, auth_user %r", targetindex,
                            auth_user)
             if not request.has_permission("pypi_submit", context=target_stage):
-               apireturn(401, message="user %r cannot upload to %r"
+                apireturn(401, message="user %r cannot upload to %r"
                                       %(auth_user, targetindex))
             self._set_versiondata_dict(target_stage, metadata)
             results.append((200, "register", name, version,
@@ -735,10 +793,15 @@ class PyPIView:
             pypiauth = (username, password)
             # prepare metadata for submission
             metadata[":action"] = "submit"
-            metadata["metadata_version"] = "1.2"
+            metadata["metadata_version"] = "2.1"
             self.log.info("registering %s-%s to %s", name, version, posturl)
             session = new_requests_session(agent=("server", server_version))
-            r = session.post(posturl, data=metadata, auth=pypiauth)
+            try:
+                r = session.post(posturl, data=metadata, auth=pypiauth)
+            except Exception as e:
+                exc_msg = ''.join(traceback.format_exception_only(e.__class__, e))
+                results.append((-1, "exception on register:", exc_msg))
+                apireturn(502, result=results, type="actionlog")
             self.log.debug("register returned: %s", r.status_code)
             results.append((r.status_code, "register", name, version))
             ok_codes = (200, 201, 410)
@@ -756,21 +819,31 @@ class PyPIView:
                     content = entry.file_get_content()
                     self.log.info("sending %s to %s, metadata %s",
                              basename, posturl, file_metadata)
-                    r = session.post(posturl, data=file_metadata,
-                          auth=pypiauth,
-                          files={"content": (basename, content)})
-                    self.log.debug("send finished, status: %s", r.status_code)
-                    results.append((r.status_code, "upload", entry.relpath,
-                                    r.text))
+                    try:
+                        r = session.post(
+                            posturl, data=file_metadata, auth=pypiauth,
+                            files={"content": (basename, content)})
+                    except Exception as e:
+                        exc_msg = ''.join(traceback.format_exception_only(e.__class__, e))
+                        results.append((-1, "exception on release upload:", exc_msg))
+                    else:
+                        self.log.debug("send finished, status: %s", r.status_code)
+                        results.append((r.status_code, "upload", entry.relpath,
+                                        r.text))
                 if links["doczip"]:
                     doc_metadata = metadata.copy()
                     doc_metadata[":action"] = "doc_upload"
                     doczip = links["doczip"][0].entry.file_get_content()
-                    r = session.post(posturl, data=doc_metadata,
-                          auth=pypiauth,
-                          files={"content": (name + ".zip", doczip)})
-                    self.log.debug("send finished, status: %s", r.status_code)
-                    results.append((r.status_code, "docfile", name))
+                    try:
+                        r = session.post(
+                            posturl, data=doc_metadata, auth=pypiauth,
+                            files={"content": (name + ".zip", doczip)})
+                    except Exception as e:
+                        exc_msg = ''.join(traceback.format_exception_only(e.__class__, e))
+                        results.append((-1, "exception on documentation upload:", exc_msg))
+                    else:
+                        self.log.debug("send finished, status: %s", r.status_code)
+                        results.append((r.status_code, "docfile", name))
                 #
             if r.status_code in ok_codes:
                 apireturn(200, result=results, type="actionlog")
@@ -823,16 +896,15 @@ class PyPIView:
                 dst=target_stage.name)
             yield (200, "store_doczip", name, version,
                    "->", target_stage.name)
-            break # we only can have one doczip for now
+            break  # we only can have one doczip for now
 
     @view_config(
         route_name="/{user}/{index}/", request_method="POST")
     def submit(self):
         request = self.request
-        context = self.context
-        if context.username == "root" and context.index == "pypi":
-            abort_submit(request, 404, "cannot submit to pypi mirror")
         stage = self.context.stage
+        if not hasattr(stage, "store_releasefile"):
+            abort_submit(request, 404, "cannot submit to mirror index")
         if not request.has_permission("pypi_submit"):
             # if there is no authenticated user, then issue a basic auth challenge
             if not request.authenticated_userid:
@@ -845,6 +917,7 @@ class PyPIView:
         except KeyError:
             abort_submit(request, 400, ":action field not found")
         if action == "submit":
+            # register always overwrites the metadata
             self._set_versiondata_form(stage, request.POST)
             return Response("")
         elif action in ("doc_upload", "file_upload"):
@@ -852,20 +925,12 @@ class PyPIView:
                 content = request.POST["content"]
             except KeyError:
                 abort_submit(request, 400, "content file field not found")
-            name = ensure_unicode(request.POST.get("name"))
+            name = ensure_unicode(request.POST["name"])
             # version may be empty on plain doczip uploads
             version = ensure_unicode(request.POST.get("version") or "")
             project = normalize_name(name)
 
             if action == "file_upload":
-                if not stage.has_project(name):
-                    abort_submit(
-                        request, 400,
-                        "no project named %r was ever registered" % (name))
-
-                self.log.debug("metadata in form: %s",
-                               list(request.POST.items()))
-
                 # we only check for release files if version is
                 # contained in the filename because for doczip files
                 # we construct the filename ourselves anyway.
@@ -876,13 +941,7 @@ class PyPIView:
                             content.filename, version))
 
                 abort_if_invalid_filename(request, name, content.filename)
-                metadata = stage.get_versiondata_perstage(project, version)
-                if not metadata:
-                    self._set_versiondata_form(stage, request.POST)
-                    metadata = stage.get_versiondata(project, version)
-                    if not metadata:
-                        abort_submit(
-                            request, 400, "could not process form metadata")
+                self._update_versiondata_form(stage, request.POST)
                 file_content = content.file.read()
                 try:
                     link = stage.store_releasefile(
@@ -898,8 +957,6 @@ class PyPIView:
                         request, 409,
                         "%s already exists in non-volatile index" % (
                             content.filename,))
-                link.add_log(
-                    'upload', request.authenticated_userid, dst=stage.name)
                 try:
                     self.xom.config.hook.devpiserver_on_upload_sync(
                         log=request.log, application_url=request.application_url,
@@ -909,6 +966,8 @@ class PyPIView:
                         request, 200,
                         "OK, but a trigger plugin failed: %s" % e, level="warn")
             else:
+                if "version" in request.POST:
+                    self._update_versiondata_form(stage, request.POST)
                 doczip = content.file.read()
                 try:
                     link = stage.store_doczip(project, version, doczip)
@@ -916,10 +975,6 @@ class PyPIView:
                     abort_submit(
                         request, 400,
                         "%s" % e)
-                except stage.MissesRegistration:
-                    abort_submit(
-                        request, 400,
-                        "%s-%s is not registered" % (name, version))
                 except stage.NonVolatile as e:
                     if e.link.matches_checksum(doczip):
                         abort_submit(
@@ -930,16 +985,18 @@ class PyPIView:
                         request, 409,
                         "%s already exists in non-volatile index" % (
                             content.filename,))
-                link.add_log(
-                    'upload', request.authenticated_userid, dst=stage.name)
+            link.add_log(
+                'upload', request.authenticated_userid, dst=stage.name)
         else:
             abort_submit(request, 400, "action %r not supported" % action)
         return Response("")
 
-    def _set_versiondata_form(self, stage, form):
+    def _get_versiondata_from_form(self, stage, form, skip_missing=False):
         metadata = {}
         for key in stage.metadata_keys:
-            if key.lower() in stage.metadata_list_fields:
+            if skip_missing and key not in form:
+                continue
+            elif key in stage.metadata_list_fields:
                 val = [ensure_unicode(item)
                         for item in form.getall(key) if item]
             else:
@@ -948,7 +1005,26 @@ class PyPIView:
                     val = ""
                 assert py.builtin._istext(val), val
             metadata[key] = val
+        return metadata
 
+    def _update_versiondata_form(self, stage, form):
+        # first we only get metadata without any default values for
+        # missing keys
+        new_metadata = self._get_versiondata_from_form(
+            stage, form, skip_missing=True)
+        # now we have the name and version and look for existing metadata
+        metadata = stage.get_versiondata_perstage(
+            new_metadata["name"], new_metadata["version"], readonly=False)
+        if not metadata:
+            # there is no existing metadata, so we do a full register
+            # with default values for missing metadata
+            new_metadata = self._get_versiondata_from_form(stage, form)
+        # finally we update and set the metadata
+        metadata.update(new_metadata)
+        self._set_versiondata_dict(stage, metadata)
+
+    def _set_versiondata_form(self, stage, form):
+        metadata = self._get_versiondata_from_form(stage, form)
         self._set_versiondata_dict(stage, metadata)
 
     def _set_versiondata_dict(self, stage, metadata):
@@ -982,11 +1058,14 @@ class PyPIView:
         if not json_preferred(self.request):
             apireturn(415, "unsupported media type %s" %
                       self.request.headers.items())
+        perstage = 'ignore_bases' in self.request.GET
         context = self.context
         view_metadata = {}
-        for version in context.list_versions():
-            view_metadata[version] = self._make_view_verdata(
-                context.get_versiondata(version=version))
+        versions = context.list_versions(perstage=perstage)
+        for version in versions:
+            versiondata = context.get_versiondata(
+                version=version, perstage=perstage)
+            view_metadata[version] = self._make_view_verdata(versiondata)
         apireturn(200, type="projectconfig", result=view_metadata)
 
     @view_config(
@@ -1063,12 +1142,23 @@ class PyPIView:
             entry_data = get_mutable_deepcopy(entry.meta)
             apireturn(200, type="releasefilemeta", result=entry_data)
 
+        if entry.last_modified is None or not entry.file_exists():
+            # The file is in a mirror and either deleted or not yet downloaded.
+            # We check whether we should serve the file directly
+            # or redirect to the external URL
+            stage = self.xom.model.getstage(
+                entry.key.params['user'],
+                entry.key.params['index'])
+            if stage is not None and stage.use_external_url:
+                # redirect to external url
+                return HTTPFound(location=entry.url)
+
         try:
             if should_fetch_remote_file(entry, request.headers):
                 app_iter = iter_fetch_remote_file(self.xom, entry)
                 headers = next(app_iter)
                 return Response(app_iter=app_iter, headers=headers)
-        except entry.BadGateway as e:
+        except BadGateway as e:
             if e.code == 404:
                 return apireturn(404, e.args[0])
             return apireturn(502, e.args[0])
@@ -1125,7 +1215,10 @@ class PyPIView:
     def index_get(self):
         stage = self.context.stage
         result = dict(stage.ixconfig)
-        result['projects'] = sorted(stage.list_projects_perstage())
+        # double negation :(
+        add_projects = 'no_projects' not in self.request.GET
+        if add_projects:
+            result['projects'] = sorted(stage.list_projects_perstage())
         apireturn(200, type="indexconfig", result=result)
 
     #
@@ -1224,25 +1317,170 @@ class PyPIView:
 
 
 def should_fetch_remote_file(entry, headers):
-    from .replica import H_REPLICA_FILEREPL
     should_fetch = not entry.file_exists()
-    # if we are asked for an "egg" development link we cause
-    # refetching it unless we are called within file replication context
-    if entry.eggfragment and not headers.get(H_REPLICA_FILEREPL):
-        should_fetch = True
     return should_fetch
 
 
+def _headers_from_response(r):
+    headers = {
+        str("X-Accel-Buffering"): str("no"),  # disable buffering in nginx
+        str("content-type"): r.headers.get(
+            "content-type", (str("application/octet-stream"), None))[0]}
+    if "last-modified" in r.headers:
+        headers[str("last-modified")] = r.headers["last-modified"]
+    if "content-length" in r.headers:
+        headers[str("content-length")] = str(r.headers["content-length"])
+    return headers
+
+
+def iter_cache_remote_file(xom, entry):
+    # we get and cache the file and some http headers from remote
+    r = xom.httpget(entry.url, allow_redirects=True)
+    if r.status_code != 200:
+        msg = "error %s getting %s" % (r.status_code, entry.url)
+        threadlog.error(msg)
+        raise BadGateway(msg, code=r.status_code, url=entry.url)
+    threadlog.info("reading remote: %s, target %s", r.url, entry.relpath)
+    content_size = r.headers.get("content-length")
+    err = None
+
+    yield _headers_from_response(r)
+
+    content = BytesIO()
+    while 1:
+        data = r.raw.read(10240)
+        if not data:
+            break
+        content.write(data)
+        yield data
+
+    content = content.getvalue()
+
+    filesize = len(content)
+    if content_size and int(content_size) != filesize:
+        err = ValueError(
+            "%s: got %s bytes of %r from remote, expected %s" % (
+                entry.relpath, filesize, r.url, content_size))
+    if not err:
+        err = entry.check_checksum(content)
+
+    if err is not None:
+        threadlog.error(str(err))
+        raise err
+
+    try:
+        # when pushing from a mirror to an index, we are still in a
+        # transaction
+        tx = entry.tx
+    except AttributeError:
+        # when streaming we won't be in a transaction anymore, so we need
+        # to open a new one below
+        tx = None
+
+    def set_content():
+        entry.file_set_content(content, r.headers.get("last-modified", None))
+        if entry.project:
+            stage = xom.model.getstage(
+                entry.key.params['user'],
+                entry.key.params['index'])
+            # for mirror indexes this makes sure the project is in the database
+            # as soon as a file was fetched
+            stage.add_project_name(entry.project)
+
+    if not entry.has_existing_metadata():
+        if tx is not None:
+            set_content()
+        else:
+            with xom.keyfs.transaction(write=True):
+                set_content()
+    else:
+        # the file was downloaded before but locally removed, so put
+        # it back in place without creating a new serial
+        # we need a direct write connection to use the io_file_* methods
+        if tx is not None:
+            tx.conn.io_file_set(entry._storepath, content)
+            threadlog.debug(
+                "put missing file back into place: %s", entry._storepath)
+        else:
+            with xom.keyfs._storage.get_connection(write=True) as conn:
+                conn.io_file_set(entry._storepath, content)
+                threadlog.debug(
+                    "put missing file back into place: %s", entry._storepath)
+                conn.commit_files_without_increasing_serial()
+
+
+def iter_remote_file_replica(xom, entry):
+    from .replica import H_REPLICA_FILEREPL, ReplicationErrors
+    replication_errors = ReplicationErrors(xom.config.serverdir)
+    # construct master URL with param
+    url = xom.config.master_url.joinpath(entry.relpath).url
+    if not entry.url:
+        threadlog.warn("missing private file: %s" % entry.relpath)
+    else:
+        threadlog.info("replica doesn't have file: %s", entry.relpath)
+    r = xom.httpget(
+        url, allow_redirects=True,
+        extra_headers={H_REPLICA_FILEREPL: str("YES")})
+    if r.status_code != 200:
+        msg = "%s: received %s from master" % (url, r.status_code)
+        if not entry.url:
+            threadlog.error(msg)
+            raise BadGateway(msg)
+        # try to get from original location
+        r = xom.httpget(entry.url, allow_redirects=True)
+        if r.status_code != 200:
+            msg = "%s\n%s: received %s" % (msg, entry.url, r.status_code)
+            threadlog.error(msg)
+            raise BadGateway(msg)
+
+    yield _headers_from_response(r)
+
+    content = BytesIO()
+    while 1:
+        data = r.raw.read(10240)
+        if not data:
+            break
+        content.write(data)
+        yield data
+
+    content = content.getvalue()
+
+    err = entry.check_checksum(content)
+    if err:
+        # the file we got is different, so we fail
+        raise BadGateway(str(err))
+
+    try:
+        # there is no code path that still has a transaction at this point,
+        # but we handle that case just to be safe
+        tx = entry.tx
+    except AttributeError:
+        # when streaming we won't be in a transaction anymore, so we need
+        # to open a new one below
+        tx = None
+    if tx is not None and tx.write:
+        entry.tx.conn.io_file_set(entry._storepath, content)
+    else:
+        # we need a direct write connection to use the io_file_* methods
+        with xom.keyfs._storage.get_connection(write=True) as conn:
+            conn.io_file_set(entry._storepath, content)
+            threadlog.debug(
+                "put missing file back into place: %s", entry._storepath)
+            conn.commit_files_without_increasing_serial()
+    # in case there were errors before, we can now remove them
+    replication_errors.remove(entry)
+
+
 def iter_fetch_remote_file(xom, entry):
     filestore = xom.filestore
     keyfs = xom.keyfs
     if not xom.is_replica():
         keyfs.restart_as_write_transaction()
         entry = filestore.get_file_entry(entry.relpath, readonly=False)
-        for part in entry.iter_cache_remote_file():
+        for part in iter_cache_remote_file(xom, entry):
             yield part
     else:
-        for part in entry.iter_remote_file_replica():
+        for part in iter_remote_file_replica(xom, entry):
             yield part
 
 
@@ -1293,13 +1531,3 @@ def abort_if_invalid_project(request, project):
             project.encode("ascii")
     except (UnicodeEncodeError, UnicodeDecodeError):
         abort(request, 400, "unicode project names not allowed")
-
-
-def getkvdict_index(hook, req):
-    kvdict = {}
-    ixconfigattrs = get_ixconfigattrs(hook, req.get("type", "stage"))
-    for key in ixconfigattrs:
-        if key in req:
-            kvdict[key] = req[key]
-    return kvdict
-
