@@ -62,6 +62,64 @@ class TestKeyFS:
         with pytest.raises(ValueError):
             pkey(hello="this/that")
 
+    @notransaction
+    def test_remove_dict_key(self, keyfs):
+        key = keyfs.add_key("NAME", "somekey", dict)
+        with keyfs.transaction(write=True) as tx:
+            tx.set(key, {u'foo': u'bar', u'ham': u'egg'})
+        with keyfs.transaction(write=True) as tx:
+            tx.set(key, {u'foo': u'bar'})
+        with keyfs.transaction(write=False) as tx:
+            assert tx.at_serial == 1
+            assert tx.get(key) == {u'foo': u'bar'}
+        with keyfs.transaction(write=False, at_serial=0) as tx:
+            assert tx.at_serial == 0
+            assert tx.get(key) == {u'foo': u'bar', u'ham': u'egg'}
+
+    @notransaction
+    def test_remove_set_item(self, keyfs):
+        key = keyfs.add_key("NAME", "somekey", set)
+        with keyfs.transaction(write=True) as tx:
+            tx.set(key, {'bar', 'egg'})
+        with keyfs.transaction(write=True) as tx:
+            tx.set(key, {'bar'})
+        with keyfs.transaction(write=False) as tx:
+            assert tx.at_serial == 1
+            assert tx.get(key) == {'bar'}
+        with keyfs.transaction(write=False, at_serial=0) as tx:
+            assert tx.at_serial == 0
+            assert tx.get(key) == {'bar', 'egg'}
+
+    @notransaction
+    def test_double_set(self, keyfs):
+        key = keyfs.add_key("NAME", "somekey", dict)
+        with keyfs.transaction(write=True) as tx:
+            tx.set(key, {u'foo': u'bar', u'ham': u'egg'})
+        with keyfs.transaction(write=True) as tx:
+            # set different value
+            tx.set(key, {})
+            # set to same value again in same transaction
+            tx.set(key, {u'foo': u'bar', u'ham': u'egg'})
+        with keyfs.transaction(write=False) as tx:
+            # the serial shouldn't have increased
+            assert tx.at_serial == 0
+            assert tx.get(key) == {u'foo': u'bar', u'ham': u'egg'}
+
+    @notransaction
+    @pytest.mark.parametrize("before,after", [
+        ({u'a': 1}, {u'b': 2}),
+        (set([3]), set([4])),
+        (5, 6)])
+    def test_delete_and_readd(self, keyfs, before, after):
+        key = keyfs.add_key("NAME", "somekey", type(before))
+        with keyfs.transaction(write=True) as tx:
+            tx.set(key, before)
+        with keyfs.transaction(write=True) as tx:
+            tx.delete(key)
+        with keyfs.transaction(write=True) as tx:
+            tx.set(key, after)
+        with keyfs.transaction(write=False) as tx:
+            assert tx.get(key) == after
 
 
 class TestGetKey:
@@ -131,9 +189,9 @@ class Test_addkey_combinations:
 
 class TestKey:
     def test_addkey_type_mismatch(self, keyfs):
-        dictkey = keyfs.add_key("NAME", "some", dict)
+        dictkey = keyfs.add_key("NAME1", "some", dict)
         pytest.raises(TypeError, lambda: dictkey.set("hello"))
-        dictkey = keyfs.add_key("NAME", "{that}/some", dict)
+        dictkey = keyfs.add_key("NAME2", "{that}/some", dict)
         pytest.raises(TypeError, lambda: dictkey(that="t").set("hello"))
 
     def test_addkey_registered(self, keyfs):
@@ -171,8 +229,8 @@ class TestKey:
         assert d2 != d
 
     def test_update(self, keyfs):
-        key1 = keyfs.add_key("NAME", "some1", dict)
-        key2 = keyfs.add_key("NAME", "some2", list)
+        key1 = keyfs.add_key("NAME1", "some1", dict)
+        key2 = keyfs.add_key("NAME2", "some2", list)
         keyfs.restart_as_write_transaction()
         with key1.update() as d:
             with key2.update() as l:
@@ -577,7 +635,7 @@ class TestSubscriber:
         pool.start()
         for i in range(10):
             with keyfs.transaction(write=True):
-                key.set(1)
+                key.set(i)
 
         l = [queue.get() for i in range(10)]
         assert sorted(l) == list(range(10))
@@ -663,7 +721,7 @@ def test_keyfs_sqlite(gentmp):
     with keyfs.transaction(write=False) as tx:
         assert tx.conn.io_file_os_path('foo') is None
         assert tx.conn.io_file_get('foo') == b'bar'
-    assert [x.basename for x in tmp.listdir()] == ['.sqlite']
+    assert [x.basename for x in tmp.listdir()] == ['.sqlite_db']
 
 
 def test_keyfs_sqlite_fs(gentmp):
